from splverifier.api.fs import FileStore, Timer, nameonly, system, rm
from splverifier.api.config import PATHS
from os import listdir as ls
from os import path
from os import mkdir
from os import getcwd
from os import environ
from shutil import copy
from shutil import move
from shutil import rmtree
from subprocess import Popen, PIPE
import sys


def introduce_dependencies(weavedir):
    """
    copies over the dependencies from the dependendy dir to weavedir
    """
    dep_dir = PATHS.autofeature.introduction_dir
    for filename in ls(dep_dir):
        src = path.join(dep_dir, filename)
        target = path.join(weavedir, filename)
        copy(src, target)


def preprocess_cfiles(srcdir, destdir):
    """
    preprocesses *.c in srcdir with gcc
    and writes to *.mc in destdir
    """
    cfiles = [f for f in ls(srcdir) if f.endswith('.c')]
    for cfile in cfiles:
        args = (path.join(srcdir, cfile), path.join(destdir, cfile[:-2]))
        system('gcc -E %s > %s.mc' % args)

def translate_specifications(specifications, srcdir, destdir):
    """
    translates *.spec in srcdir using cpawsl and
    writes to *_spec.c in destdir
    These files are then preprocessed
    and written to *_spec.acc
    """
    
    for specification in specifications:
        spec_name = nameonly(specification)
        abs_destfile = path.join(destdir, spec_name + "_spec.c")
        abs_destfile_acc = path.join(destdir, spec_name + "_spec.acc")
        args = ('java', '-jar', PATHS.autofeature.autofeature_jar, specification)
        stdout, stderr = Popen(args, stdout=PIPE, stderr=PIPE).communicate()
        if stderr:
            print "Error while translating %s :" % specification
            print
            print stderr
            print
            print "Skipping %s!" % specification
        else:
            outf = open(abs_destfile, 'w')
            outf.write(stdout)
            outf.close()
            system("gcc -I%s -E %s > %s" % (srcdir, abs_destfile, abs_destfile_acc))
            rm(abs_destfile)

def translate_acc_file(processed_dir, specs):
    """
    runs acc on preprocessed sources and the generated acc files
    """
    system("%s %s/*.mc %s" % (PATHS.autofeature.acc_bin, processed_dir, ' '.join(specs)))

def translate_acc_files(processed_dir, target_dir):
    """
    creates a dir under spec_build for each spec 
    runs acc to weave C and acc for all specs
    """
    processed_fs = FileStore(processed_dir)
    acc_specs = processed_fs.glob('*_spec.acc')
    c_files = [x[:-3] + '.c' for x in ls(processed_dir) if x.endswith('.mc')]
    
    translate_acc_file(processed_dir, acc_specs)
    
    for aspect in processed_fs.glob('*_spec.c'):
        move(aspect, target_dir)
    for c_file in c_files:
        move(path.join(processed_dir, c_file), target_dir)
        


def get_selected_features(expressionfile):
    """
    extract selected features from features.exp
    """
    f = open(expressionfile, 'r')
    features = [x.strip() for x in f.readlines()]
    f.close()
    return features

def copy_specfiles(feature_dir, srcdir):
    """
    copy specification files to the src-composed folder.
    The filenames are prefixed with the feature the spec came from
    to avoid naming conflicts.
    """
    features = get_selected_features(path.join(srcdir, '..', 'features.exp'))

    for feature in features:
        fdir = path.join(feature_dir, feature)
        specfiles = [x for x in ls(fdir) if x.endswith('.spec')]
        for spec in specfiles:
            outfile = path.join(srcdir, feature + '__' + spec)
            copy(path.join(fdir, spec), outfile)

def copy_main(composed_dir, scenario):
    """
    copies Test1.c and Test1.h to src-composed.
    if these are not found falls back to Test.c and Test.h.
    if these are not found nothing is copied.

    FIXME hardcoded ... find a better way to do this
    """
    fsc = path.join(composed_dir, 'featureselect.c')
    if scenario:
        srcdir = path.join(composed_dir, '..')
        prefix = scenario
        if not path.isfile(prefix + '.c'):
            raise Exception("Scenario %s.c does not exist" % scenario)
        
        for ending in ('.c', '.h'):
            src = path.join(srcdir, prefix + ending)
            if path.isfile(src):
                copy(src, path.join(composed_dir, prefix + ending))


def copyspec():
    base = '/home/henne/share/allvariants/EncryptDecrypt'
    spec = '/home/henne/share/allvariants/EncryptDecrypt/EncryptDecrypt_var.spec'
    for adir in os.listdir(base):
        if os.path.isdir(base + '/' + adir):
            local('cp %s %s/%s/' % (spec, base, adir))


def weave_specifications(composed, specifications, target_dir):
    """
    - preprocesses C files in SRC_DIR
    - translates specification files to ACC
    - invokes ACC to weave preprocessed files and translated specification files
    """

    timer = Timer().start()
    
    tmpfs = FileStore(PATHS.autofeature.tmpdir)
    tmpfs.clear()
    tmpfs.create()
    
    FileStore(composed).copycontents(tmpfs.get())
    
    introduce_dependencies(tmpfs.get())
    
    build_fs = tmpfs.get_fs('build').create()
    weave_fs = tmpfs.get_fs('weave').create()

    preprocess_cfiles(tmpfs.get(), build_fs.get())
    translate_specifications(specifications, tmpfs.get(), build_fs.get())
    translate_acc_files(build_fs.get(), weave_fs.get())
    timer.stop()
    weave_fs.moveto(target_dir)
    return timer.elapsed
    
